#
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import json
import os
from string import Template
import subprocess
import tempfile
from threading import Lock
import logging

from cocaine.tools.printer import printer
from cocaine.tools.repository import GitRepositoryDownloader, RepositoryDownloadError

__author__ = 'EvgenySafronov <division494@gmail.com>'


class ModuleInstallError(Exception):
    pass


log = logging.getLogger(__name__)


MODULE_INSTALL_START = 'Start installing python module'
MODULE_INSTALL_FINISH = 'Python module has been successfully installed'

VENV_CREATE_START = 'Creating virtual environment ...'
VENV_CREATE_FINISH = 'Virtual environment has been successfully created in "{0}"'
VENV_CREATE_ERROR = 'Module virtualenv is not installed, so a new environment cannot be created'


COCAINE_PYTHON_FRAMEWORK_URL = 'git@github.com:cocaine/cocaine-framework-python.git'
COCAINE_DOWNLOAD_START = 'Downloading cocaine-framework-python from "{0}" to "{1}"...'

COCAINE_INSTALL_START = 'Installing cocaine-framework-python ...'
COCAINE_INSTALL_FINISH = 'Cocaine-framework-python has been successfully installed'
COCAINE_INSTALL_ERROR = 'cocaine-framework-python install error'

BOOTSTRAP_FILENAME = 'bootstrap.sh'
BOOTSTRAP_TEMPLATE = '''#!/bin/sh
source ${virtualEnvironmentPath}/bin/activate
python ${slave} $$@
'''
BOOTSTRAP_CREATE_START = 'Creating bootstrap script ...'
BOOTSTRAP_CREATE_ERROR = 'bootstrap create error - {0}'

MANIFEST_FILENAME = 'manifest.json'
MANIFEST_READ_OK = 'Reading manifest file: "{0}"'
MANIFEST_SLAVE_PARSE_ERROR = 'manifest read error - slave is not specified'
MANIFEST_READ_ERROR = 'manifest read error - {0}'
MANIFEST_LOCATE_ERROR = 'locate manifest error - {0}'

REQUIREMENTS_FILENAME = 'requirements.txt'
REQUIREMENTS_NO_FILES_FOUND = 'No requirements found. Skipping this step ...'
REQUIREMENTS_INSTALL_START = 'Installing requirements ...'
REQUIREMENTS_INSTALL_FINISH = 'All requirements has been successfully installed'


class ModuleInstaller(object):
    def install(self):
        raise NotImplementedError


class PythonModuleInstaller(ModuleInstaller):
    virtualEnvGlobalLock = Lock()

    def __init__(self, path, outputPath, manifestPath=None, virtualEnvironmentName='venv', stream=None):
        self.path = path
        self.outputPath = outputPath
        self.manifestPath = manifestPath
        self.virtualEnvironmentPath = os.path.join(self.outputPath, virtualEnvironmentName)
        self.stream = stream or open(os.devnull, 'w')

        if not os.path.exists(self.path):
            raise ValueError('path "{0}" is not valid module path'.format(self.path))

        if not os.path.exists(self.outputPath):
            raise ValueError('path "{0}" is not valid path'.format(self.outputPath))

        if os.path.exists(self.virtualEnvironmentPath):
            raise ValueError('virtual environment is already exists in "{0}"'.format(self.virtualEnvironmentPath))

    def install(self):
        log.debug(MODULE_INSTALL_START)
        self.createVirtualEnvironment()
        self.prepareModule()
        self.installCocaineFramework()
        self.installRequirements()
        log.debug(MODULE_INSTALL_FINISH)

    def createVirtualEnvironment(self):
        try:
            with self.virtualEnvGlobalLock:
                import virtualenv
                log.debug(VENV_CREATE_START)
                virtualenv.create_environment(self.virtualEnvironmentPath)
                log.debug(VENV_CREATE_FINISH.format(self.outputPath))
        except ImportError:
            raise ModuleInstallError(VENV_CREATE_ERROR)

    def prepareModule(self):
        log.debug(BOOTSTRAP_CREATE_START)
        manifest = self._readManifest(self.path)
        self._createBootstrap(os.path.join(self.outputPath, BOOTSTRAP_FILENAME), manifest)
        self._copyManifest(os.path.join(self.outputPath, MANIFEST_FILENAME), manifest)

    def _readManifest(self, path):
        try:
            if not self.manifestPath:
                self.manifestPath = _locateFile(path, MANIFEST_FILENAME)

            with open(self.manifestPath) as fh:
                log.debug(MANIFEST_READ_OK.format(self.manifestPath))
                try:
                    manifest = json.loads(fh.read())
                    if 'slave' not in manifest:
                        raise ModuleInstallError(MANIFEST_SLAVE_PARSE_ERROR)
                    return manifest
                except (IOError, ValueError) as err:
                    raise ModuleInstallError(MANIFEST_READ_ERROR.format(err))
        except IOError as err:
            raise ModuleInstallError(MANIFEST_LOCATE_ERROR.format(err))

    def _createBootstrap(self, path, manifest):
        try:
            with open(path, 'w') as fh:
                fh.write(Template(BOOTSTRAP_TEMPLATE).substitute(**{
                    'virtualEnvironmentPath': self.virtualEnvironmentPath,
                    'slave': manifest['slave']
                }))
            os.chmod(path, 493)  # 0755
        except IOError as err:
            raise ModuleInstallError(BOOTSTRAP_CREATE_ERROR.format(err))

    def _copyManifest(self, path, manifest):
        try:
            with open(path, 'w') as fh:
                manifest['slave'] = BOOTSTRAP_FILENAME
                fh.write(json.dumps(manifest))
        except IOError as err:
            raise ModuleInstallError(BOOTSTRAP_CREATE_ERROR.format(err))

    def installCocaineFramework(self):
        path = tempfile.mkdtemp()
        downloader = GitRepositoryDownloader(stream=self.stream)
        try:
            log.debug(COCAINE_DOWNLOAD_START.format(COCAINE_PYTHON_FRAMEWORK_URL, path))
            downloader.download(COCAINE_PYTHON_FRAMEWORK_URL, path)
        except RepositoryDownloadError as err:
            raise ModuleInstallError(err)

        log.debug(COCAINE_INSTALL_START)
        python = os.path.join(self.virtualEnvironmentPath, 'bin', 'python')
        process = subprocess.Popen([python, 'setup.py', 'install', '--without-tools'],
                                   cwd=path,
                                   stdout=self.stream,
                                   stderr=self.stream)
        process.wait()
        if process.returncode != 0:
            raise ModuleInstallError(COCAINE_INSTALL_ERROR)
        log.debug(COCAINE_INSTALL_FINISH)

    def installRequirements(self):
        try:
            requirementsPath = _locateFile(self.path, REQUIREMENTS_FILENAME)
        except IOError:
            requirementsPath = ''

        if not requirementsPath:
            log.debug(REQUIREMENTS_NO_FILES_FOUND)
            return

        log.debug(REQUIREMENTS_INSTALL_START)
        _pip = os.path.join(self.virtualEnvironmentPath, 'bin', 'pip')
        process = subprocess.Popen([_pip, 'install', '-r', requirementsPath],
                                   stdout=self.stream,
                                   stderr=self.stream)
        process.wait()
        if process.returncode != 0:
            raise ModuleInstallError()
        log.debug(REQUIREMENTS_INSTALL_FINISH)


def _locateFile(path, filenameLocate):
        with printer('Locating %s', filenameLocate) as p:
            basename, separator, extension = filenameLocate.partition('.')
            locatedFilenames = []
            for root, dirNames, filenames in os.walk(path):
                for filename in filenames:
                    if filename.startswith(basename):
                        priority = 1
                        if root == path:
                            priority += 100
                        if filename == filenameLocate:
                            priority += 10
                        locatedFilenames.append((os.path.join(root, filename), priority))
            locatedFilenames = sorted(locatedFilenames, key=lambda filename: filename[1], reverse=True)
            log.debug('Filenames found: {0}'.format(locatedFilenames))
            if not locatedFilenames:
                raise IOError('No files found in "{0}" or subdirectories'.format(path))

            filename, priority = locatedFilenames[0]
            p('found "%s"', filename)
            return os.path.abspath(filename)



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

import collections
import logging
import os
import sys

from opster import Dispatcher

from cocaine.services import Locator, Service
from cocaine.tools import ColoredFormatter, interactiveEmit
from cocaine.tools.cli import Executor
from cocaine.tools.error import ToolsError
from cocaine.tools.version import __version__

__author__ = 'Evgeny Safronov <division494@gmail.com>'


DESCRIPTION = ''
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 10053


class Global(object):
    options = [
        ('h', 'host', DEFAULT_HOST, 'hostname'),
        ('p', 'port', DEFAULT_PORT, 'port'),
        ('', 'timeout', 20.0, 'timeout, s'),
        ('', 'color', True, 'enable colored output'),
        ('', 'debug', ('disable', 'tools', 'all'), 'enable debug mode'),
    ]

    def __init__(self, host=DEFAULT_HOST, port=DEFAULT_PORT, color=False, timeout=False, debug=False):
        self.host = host
        self.port = port
        self.endpoints = [(self.host, self.port)]
        self.timeout = timeout
        self.executor = Executor(timeout)
        self._locator = None
        self.configureLog(debug=debug, color=color)

    @staticmethod
    def configureLog(debug='disable', color=True, logNames=None):
        if not logNames:
            logNames = ['cocaine.tools']
        message = '%(message)s'
        level = logging.INFO
        if debug != 'disable':
            message = '[%(asctime)s] %(module)s %(name)s:%(lineno)d %(levelname)-8s: %(message)s'
            level = logging.DEBUG

        ch = logging.StreamHandler()
        if debug == 'disable':
            setattr(logging.StreamHandler, logging.StreamHandler.emit.__name__, interactiveEmit)
        ch.fileno = ch.stream.fileno
        ch.setLevel(level)
        formatter = ColoredFormatter(message, colored=color and sys.stdin.isatty())
        ch.setFormatter(formatter)

        if debug == 'all':
            logNames.append('cocaine')

        for logName in logNames:
            log = logging.getLogger(logName)
            log.setLevel(logging.DEBUG)
            log.propagate = False
            log.addHandler(ch)

    @property
    def locator(self):
        if self._locator:
            return self._locator
        else:
            try:
                locator = Locator(endpoints=self.endpoints)
                self._locator = locator
                return locator
            except Exception as err:
                raise ToolsError(err)

    def getService(self, name):
        try:
            service = Service(name, endpoints=self.endpoints)
            return service
        except Exception as err:
            raise ToolsError(err)


def middleware(func):
    def extract_dict(source, *keys):
        dest = {}
        for k in keys:
            dest[k] = source.pop(k, None)
        return dest

    def inner(*args, **kwargs):
        opts = extract_dict(kwargs, 'host', 'port', 'color', 'timeout', 'debug')
        if func.__name__ == 'help_inner':
            return func(*args, **kwargs)
        locator = Global(**opts)
        return func(locator, *args, **kwargs)
    return inner


d = Dispatcher(globaloptions=Global.options, middleware=middleware)
appDispatcher = Dispatcher(globaloptions=Global.options, middleware=middleware)


class dispatcher(object):
    group = Dispatcher(globaloptions=Global.options, middleware=middleware)


profileDispatcher = Dispatcher(globaloptions=Global.options, middleware=middleware)
runlistDispatcher = Dispatcher(globaloptions=Global.options, middleware=middleware)
crashlogDispatcher = Dispatcher(globaloptions=Global.options, middleware=middleware)
tracingDispatcher = Dispatcher(globaloptions=Global.options, middleware=middleware)


@d.command(name='locate', usage='[--name=NAME]')
def locate(options,
           name=('n', '', 'service name')):
    """Show information about requested service"""
    options.executor.executeAction('locate', **{
        'name': name,
        'locator': options.locator,
    })


@d.command(name='version')
def show_version(dummy_options):
    """Show version"""
    print(__version__)


@d.command(name='routing', usage='[--name=NAME]')
def routing(options,
            name=('n', '', 'group name')):
    """Show information about the requested routing group"""
    options.executor.executeAction('routing', **{
        'name': name,
        'locator': options.locator,
    })


@d.command(name='cluster')
def cluster(options,
            resolve=('r', False, 'show IPs instead of hostnames')):
    """Show cluster info"""
    options.executor.executeAction('cluster', **{
        'locator': options.getService('locator'),
        # Actually we have IPs and we need not do anything
        # to resolve them to IPs.
        # So the default behavior fits better to this option name.
        'resolve': resolve,
    })


@d.command(name='info', usage='[--name=NAME] [-pmb]')
def info(options,
         name=('n', '', 'application name'),
         profile=('p', False, 'expand profile'),
         manifest=('m', False, 'expand manifest'),
         brief=('b', False, 'show brief info only (disable -p and -m)'),
         no_wildcard=('w', True, 'do not use wildcard to match app name'),
         info_timeout=('t', 5., 'wait timeout in sec for each app')):
    """Show information about cocaine runtime

    Return json-like string with information about cocaine-runtime.

    If the name option is not specified, shows information about all applications.
    Flags can be specified to set detalization of the output.
    """
    flag_brief = 0x00
    flag_verbose = 0x01
    flag_manifest = 0x02
    flag_profile = 0x04
    # bried disables all further flags
    flags = flag_brief if brief else flag_verbose

    if manifest:
        flags |= flag_manifest
    if profile:
        flags |= flag_profile

    options.executor.executeAction('info', **{
        'node': options.getService('node'),
        'locator': options.locator,
        'name': name,
        'flags': flags,
        'use_wildcard': no_wildcard,
        'timeout': info_timeout,
    })


@appDispatcher.command(name='list')
def app_list(options):
    """Show installed applications list."""
    options.executor.executeAction('app:list', **{
        'storage': options.getService('storage')
    })


@appDispatcher.command(usage='--name=NAME', name='view')
def app_view(options,
             name=('n', '', 'application name')):
    """Show manifest context for application.

    If application is not uploaded, an error will be displayed.
    """
    options.executor.executeAction('app:view', **{
        'storage': options.getService('storage'),
        'name': name,
    })


@appDispatcher.command(name='upload', usage='[PATH] [--name=NAME] [--manifest=MANIFEST] [--package=PACKAGE]')
def app_upload(options,
               path=None,
               name=('n', '', 'application name'),
               manifest=('', '', 'manifest file name'),
               package=('', '', 'path to the application archive'),
               docker_address=('', '', 'docker address'),
               registry=('', '', 'registry address'),
               recipe=('', '', 'path to the recipe file'),
               manifest_only=('', False, 'upload manifest only')):
    """Upload application with its environment (directory) into the storage.

    Application directory or its subdirectories must contain valid manifest file named `manifest.json` or `manifest`
    otherwise you must specify it explicitly by setting `--manifest` option.

    You can specify application name. By default, leaf directory name is treated as application name.

    If you have already prepared application archive (*.tar.gz), you can explicitly specify path to it by setting
    `--package` option.

    You can control process of creating and uploading application by specifying `--debug=tools` option. This is helpful
    when some errors occurred.
    """
    TIMEOUT_THRESHOLD = 120.0
    if options.executor.timeout < TIMEOUT_THRESHOLD:
        logging.getLogger('cocaine.tools').info('Setting timeout to the %fs', TIMEOUT_THRESHOLD)
        options.executor.timeout = TIMEOUT_THRESHOLD
    MutexRecord = collections.namedtuple('MutexRecord', 'value, name')
    mutex = [
        (MutexRecord(path, 'PATH'), MutexRecord(package, '--package')),
        (MutexRecord(package, '--package'), MutexRecord(docker_address, '--docker')),
        (MutexRecord(package, '--package'), MutexRecord(registry, '--registry')),
    ]
    for (f, s) in mutex:
        if f.value and s.value:
            print('Wrong usage: option {0} and {1} are mutual exclusive, you can only force one'.format(f.name, s.name))
            exit(os.EX_USAGE)

    if manifest_only:
        options.executor.executeAction('app:upload-manual', **{
            'storage': options.getService('storage'),
            'name': name,
            'manifest': manifest,
            'package': None,
            'manifest_only': manifest_only,
        })
    elif package:
        options.executor.executeAction('app:upload-manual', **{
            'storage': options.getService('storage'),
            'name': name,
            'manifest': manifest,
            'package': package
        })
    elif docker_address:
        options.executor.executeAction('app:upload-docker', **{
            'storage': options.getService('storage'),
            'path': path,
            'name': name,
            'manifest': manifest,
            'address': docker_address,
            'registry': registry
        })
    else:
        options.executor.executeAction('app:upload', **{
            'storage': options.getService('storage'),
            'path': path,
            'name': name,
            'manifest': manifest
        })


@appDispatcher.command(name='import', usage='[PATH] [--name=NAME] [--manifest=MANIFEST] [--package=PACKAGE]')
def app_import(options,
               path=None,
               name=('n', '', 'application name'),
               manifest=('', '', 'manifest file name'),
               container_url=('', '', 'docker container url'),
               docker_address=('', '', 'docker address'),
               registry=('', '', 'registry address'),
               recipe=('', '', 'path to the recipe file'),
               manifest_only=('', False, 'upload manifest only')):
    """Import application's docker container

    You can control process of creating and uploading application by specifying `--debug=tools` option. This is helpful
    when some errors occurred.
    """
    timeout_threshold = 120.0
    if options.executor.timeout < timeout_threshold:
        logging.getLogger('cocaine.tools').info('Setting timeout to the %fs', timeout_threshold)
        options.executor.timeout = timeout_threshold

    if container_url and docker_address:
        options.executor.executeAction('app:import-docker', **{
            'storage': options.getService('storage'),
            'path': path,
            'name': name,
            'manifest': manifest,
            'container': container_url,
            'address': docker_address,
            'registry': registry
        })
    else:
        print("wrong usage")
        exit(os.EX_USAGE)


@appDispatcher.command(name='remove')
def app_remove(options,
               name=('n', '', 'application name')):
    """Remove application from storage.

    No error messages will display if specified application is not uploaded.
    """
    options.executor.executeAction('app:remove', **{
        'storage': options.getService('storage'),
        'name': name
    })


@appDispatcher.command(name='start')
def app_start(options,
              name=('n', '', 'application name'),
              profile=('r', '', 'profile name')):
    """Start application with specified profile.

    Does nothing if application is already running.
    """
    options.executor.executeAction('app:start', **{
        'node': options.getService('node'),
        'name': name,
        'profile': profile
    })


@appDispatcher.command(name='pause')
def app_pause(options,
              name=('n', '', 'application name')):
    """Stop application.

    This command is alias for ```cocaine-tool app stop```.
    """
    options.executor.executeAction('app:pause', **{
        'node': options.getService('node'),
        'name': name
    })


@appDispatcher.command(name='stop')
def app_stop(options,
             name=('n', '', 'application name')):
    """Stop application."""
    options.executor.executeAction('app:stop', **{
        'node': options.getService('node'),
        'name': name
    })


@appDispatcher.command(name='restart')
def app_restart(options,
                name=('n', '', 'application name'),
                profile=('r', '', 'profile name')):
    """Restart application.

    Executes ```cocaine-tool app pause``` and ```cocaine-tool app start``` sequentially.

    It can be used to quickly change application profile.
    """
    options.executor.executeAction('app:restart', **{
        'node': options.getService('node'),
        'locator': options.locator,
        'name': name,
        'profile': profile
    })


@appDispatcher.command()
def check(options,
          name=('n', '', 'application name')):
    """Checks application status."""
    options.executor.executeAction('app:check', **{
        'node': options.getService('node'),
        'storage': options.getService('storage'),
        'locator': options.locator,
        'name': name,
    })


@profileDispatcher.command(name='list')
def profile_list(options):
    """Show installed profiles."""
    options.executor.executeAction('profile:list', **{
        'storage': options.getService('storage')
    })


@profileDispatcher.command(name='view')
def profile_view(options,
                 name=('n', '', 'profile name')):
    """Show profile configuration context."""
    options.executor.executeAction('profile:view', **{
        'storage': options.getService('storage'),
        'name': name
    })


@profileDispatcher.command(name='upload')
def profile_upload(options,
                   name=('n', '', 'profile name'),
                   profile=('', '', 'path to profile file')):
    """Upload profile into the storage."""
    options.executor.executeAction('profile:upload', **{
        'storage': options.getService('storage'),
        'name': name,
        'profile': profile
    })


@profileDispatcher.command(name='edit')
def profile_edit(options,
                 name=('n', '', 'profile name')):
    """Edit profile in interactive editor."""
    options.executor.timeout = None
    options.executor.executeAction('profile:edit', **{
        'storage': options.getService('storage'),
        'name': name,
    })


@profileDispatcher.command(name='remove')
def profile_remove(options,
                   name=('n', '', 'profile name')):
    """Remove profile from the storage."""
    options.executor.executeAction('profile:remove', **{
        'storage': options.getService('storage'),
        'name': name
    })


@profileDispatcher.command(name='copy')
def profile_copy(options,
                 name=('n', '', 'profile name'),
                 copyname=('c', '', 'new profile name')):
    """Copy a profile"""
    options.executor.executeAction('profile:copy', **{
        'storage': options.getService('storage'),
        'name': name,
        'copyname': copyname,
    })


@profileDispatcher.command(name='rename')
def profile_rename(options,
                   name=('n', '', 'profile name'),
                   copyname=('c', '', 'new profile name')):
    """Raname a profile"""
    options.executor.executeAction('profile:rename', **{
        'storage': options.getService('storage'),
        'name': name,
        'copyname': copyname,
    })


@runlistDispatcher.command(name='list')
def runlist_list(options):
    """Show uploaded runlists."""
    options.executor.executeAction('runlist:list', **{
        'storage': options.getService('storage')
    })


@runlistDispatcher.command(name='view')
def runlist_view(options,
                 name=('n', '', 'name')):
    """Show configuration context for runlist."""
    options.executor.executeAction('runlist:view', **{
        'storage': options.getService('storage'),
        'name': name
    })


@runlistDispatcher.command(name='edit', usage='NAME')
def runlist_edit(options,
                 name=('n', '', 'runlist name')):
    """Edit runlist interactively."""
    options.executor.timeout = None
    options.executor.executeAction('runlist:edit', **{
        'storage': options.getService('storage'),
        'name': name
    })


@runlistDispatcher.command(name='upload')
def runlist_upload(options,
                   name=('n', '', 'name'),
                   runlist=('', '', 'path to the runlist configuration json file')):
    """Upload runlist with context into the storage."""
    options.executor.executeAction('runlist:upload', **{
        'storage': options.getService('storage'),
        'name': name,
        'runlist': runlist
    })


@runlistDispatcher.command(name='create')
def runlist_create(options,
                   name=('n', '', 'name')):
    """Create runlist and upload it into the storage."""
    options.executor.executeAction('runlist:create', **{
        'storage': options.getService('storage'),
        'name': name
    })


@runlistDispatcher.command(name='copy', usage='--name=NAME --copyname=NEWNAME')
def runlist_copy(options,
                 name=('n', '', 'name'),
                 copyname=('c', '', 'copyname')):
    """Copy runlist."""
    options.executor.executeAction('runlist:copy', **{
        'storage': options.getService('storage'),
        'name': name,
        'copyname': copyname,
    })


@runlistDispatcher.command(name='remove')
def runlist_remove(options,
                   name=('n', '', 'name')):
    """Remove runlist from the storage."""
    options.executor.executeAction('runlist:remove', **{
        'storage': options.getService('storage'),
        'name': name
    })


@runlistDispatcher.command(name='rename', usage='--name=NAME --copyname=NEWNAME')
def runlist_rename(options,
                   name=('n', '', 'name'),
                   copyname=('c', '', 'copyname')):
    """Rename runlist."""
    options.executor.executeAction('runlist:rename', **{
        'storage': options.getService('storage'),
        'name': name,
        'copyname': copyname,
    })


@runlistDispatcher.command(name='add-app')
def runlist_add_app(options,
                    name=('n', '', 'runlist name'),
                    app=('', '', 'application name'),
                    profile=('', '', 'suggested profile'),
                    force=('', False, 'create runlist if it is not exist')):
    """Add specified application with profile to the runlist.

    Existence of application or profile is not checked.
    """
    options.executor.executeAction('runlist:add-app', **{
        'storage': options.getService('storage'),
        'name': name,
        'app': app,
        'profile': profile,
        'force': force
    })


@runlistDispatcher.command(name='remove-app')
def runlist_remove_app(options,
                       name=('n', '', 'runlist name'),
                       app=('', '', 'application name')):
    """Remove specified application from the runlist.
    """
    options.executor.executeAction('runlist:remove-app', **{
        'storage': options.getService('storage'),
        'name': name,
        'app': app
    })


@crashlogDispatcher.command(name='status')
def crashlog_status(options):
    """Show crashlog status.
    """
    options.executor.executeAction('crashlog:status', **{
        'storage': options.getService('storage'),
    })


@crashlogDispatcher.command(name='list')
def crashlog_list(options,
                  name=('n', '', 'name'),
                  day=('d', '', 'day')):
    """Show crashlogs list for application.

    Prints crashlog list in timestamp - uuid format.
    """
    options.executor.executeAction('crashlog:list', **{
        'storage': options.getService('storage'),
        'name': name,
        'day_string': day,
    })


@crashlogDispatcher.command(name='view')
def crashlog_view(options,
                  name=('n', '', 'name'),
                  timestamp=('t', '', 'timestamp')):
    """Show crashlog for application with specified timestamp."""
    options.executor.executeAction('crashlog:view', **{
        'storage': options.getService('storage'),
        'name': name,
        'timestamp': timestamp
    })


@crashlogDispatcher.command(name='remove')
def crashlog_remove(options,
                    name=('n', '', 'name'),
                    timestamp=('t', '', 'timestamp')):
    """Remove crashlog for application with specified timestamp from the storage."""
    options.executor.executeAction('crashlog:remove', **{
        'storage': options.getService('storage'),
        'name': name,
        'timestamp': timestamp
    })


@crashlogDispatcher.command(name='removeall')
def crashlog_removeall(options,
                       name=('n', '', 'name')):
    """Remove all crashlogs for application from the storage."""
    options.executor.executeAction('crashlog:removeall', **{
        'storage': options.getService('storage'),
        'name': name,
    })


@crashlogDispatcher.command(name='clean')
def crashlog_clean(options,
                   name=('n', '', 'name'),
                   timestamp=('t', '', 'timestamp'),
                   size=('s', 1000, 'size')):
    """For application [NAME] leave [SIZE] crashlogs or remove all crashlogs with timestamp > [TIMESTAMP]."""
    options.executor.executeAction('crashlog:clean', **{
        'storage': options.getService('storage'),
        'name': name,
        'size': size,
        'timestamp': timestamp
    })


@crashlogDispatcher.command(name='cleanrange')
def crashlog_cleanrange(options,
                        from_day=('f', '', 'from day'),
                        up_to_day=('u', 'yesterday', 'up to day')):
    """Remove all crashlogs from one up to another.
    The date can be specified as DAY-[MONTH-[YEAR]].
    Example:
        today, yesterday,
        10, 10-09, 10-09-2015"""
    options.executor.executeAction('crashlog:cleanwhen', **{
        'storage': options.getService('storage'),
        'from_day': from_day,
        'to_day': up_to_day,
    })


@dispatcher.group.command(name='list', usage='')
def group_list(options):
    """Show routing groups.
    """
    options.executor.executeAction('group:list', **{
        'storage': options.getService('storage')
    })


@dispatcher.group.command(name='view', usage='-n NAME')
def group_view(options,
               name=('n', '', 'group name')):
    """Show specified routing group.
    """
    options.executor.executeAction('group:view', **{
        'storage': options.getService('storage'),
        'name': name
    })


@dispatcher.group.command(name='create', usage='-n NAME -c CONTENT')
def group_create(options,
                 name=('n', '', 'group name'),
                 content=('c', '', 'group content')):
    """Create routing group.
    You can optionally specify content for created routing group. It can be both direct json expression in single
    quotes, or path to the json file with settings. The settings itself must be key-value list, where `key` represents
    application name, and `value` represents its weight. For example:

    cocaine-tool group create -n new_group -c '{
        "app": 1,
        "another_app": 2
    }'.

    Warning: all application weights must be positive integers,
             total weight must be positive.
    """
    options.executor.executeAction('group:create', **{
        'storage': options.getService('storage'),
        'name': name,
        'content': content
    })


@dispatcher.group.command(name='remove', usage='--name=NAME')
def group_remove(options,
                 name=('n', '', 'group name')):
    """Remove routing group"""
    options.executor.executeAction('group:remove', **{
        'storage': options.getService('storage'),
        'name': name
    })


@dispatcher.group.command(name='copy', usage='--name=NAME --copyname=NEWNAME')
def group_copy(options,
               name=('n', '', 'group name'),
               copyname=('c', '', 'new group name')):
    """Copy routing group."""
    options.executor.executeAction('group:copy', **{
        'storage': options.getService('storage'),
        'name': name,
        'copyname': copyname,
    })


@dispatcher.group.command(name='rename', usage='--name=NAME --copyname=NEWNAME')
def group_rename(options,
                 name=('n', '', 'group name'),
                 copyname=('c', '', 'new group name')):
    """Rename routing group."""
    options.executor.executeAction('group:rename', **{
        'storage': options.getService('storage'),
        'name': name,
        'copyname': copyname,
    })


@dispatcher.group.command(name='refresh', usage='[NAME]')
def group_refresh(options,
                  name=None):
    """Refresh routing group.

    If group name is empty this command will refresh all groups.
    """
    options.executor.executeAction('group:refresh', **{
        'locator': options.locator,
        'storage': options.getService('storage'),
        'name': name
    })


@dispatcher.group.command(name='push', usage='-n=NAME --app=APP -w=WEIGHT')
def group_push(options,
               name=('n', '', 'group name'),
               app=('', '', 'app name'),
               weight=('w', '', 'weight')):
    """Add application with its weight into the routing group.

    Warning: application weight must be positive integer.
    """
    options.executor.executeAction('group:app:add', **{
        'storage': options.getService('storage'),
        'name': name,
        'app': app,
        'weight': weight
    })


@dispatcher.group.command(name='pop', usage='-n NAME --app=APP')
def group_pop(options,
              name=('n', '', 'group name'),
              app=('', '', 'app name')):
    """Remove application from routing group.
    """
    options.executor.executeAction('group:app:remove', **{
        'storage': options.getService('storage'),
        'name': name,
        'app': app
    })


@dispatcher.group.command(name='edit', usage='-n NAME')
def group_edit(options,
               name=('n', '', 'group name')):
    """Edit group in an interactive editor"""
    options.executor.timeout = None
    options.executor.executeAction('group:edit', **{
        'storage': options.getService('storage'),
        'name': name,
    })


@tracingDispatcher.command(name='store', usage='-n NAME -v VALUE')
def tracing_store(options,
                  name=('n', '', 'node name'),
                  value=('v', '', 'valuer')):
    options.executor.executeAction('tracing:store', **{
        'configuration_service': options.getService('unicorn'),
        'name': name,
        'value': value,
    })


@tracingDispatcher.command(name='remove', usage='-n NAME')
def tracing_remove(options,
                   name=('n', '', 'node name')):
    options.executor.executeAction('tracing:remove', **{
        'configuration_service': options.getService('unicorn'),
        'name': name,
    })


@tracingDispatcher.command(name='view', usage='-n NAME')
def tracing_view(options,
                 name=('n', '', 'node name')):
    options.executor.executeAction('tracing:view', **{
        'configuration_service': options.getService('unicorn'),
        'name': name,
    })

d.nest('app', appDispatcher, 'application commands')
d.nest('profile', profileDispatcher, 'profile commands')
d.nest('runlist', runlistDispatcher, 'runlist commands')
d.nest('crashlog', crashlogDispatcher, 'crashlog commands')
d.nest('group', dispatcher.group, 'routing group commands')
d.nest('tracing', tracingDispatcher, 'tracing configuration commands')



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import types

from tornado import gen

from cocaine.decorators import coroutine

from cocaine.tools.error import ToolsError
from cocaine.tools import actions, log
from cocaine.tools.actions import CocaineConfigReader
from cocaine.tools.printer import printer
from cocaine.tools.tags import RUNLISTS_TAGS

__author__ = 'Evgeny Safronov <division494@gmail.com>'


class Specific(actions.Specific):
    def __init__(self, storage, name):
        super(Specific, self).__init__(storage, 'runlist', name)


class List(actions.List):
    def __init__(self, storage):
        super(List, self).__init__('runlists', RUNLISTS_TAGS, storage)


class View(actions.View):
    def __init__(self, storage, name):
        super(View, self).__init__(storage, 'runlist', name, 'runlists')


class Upload(Specific):
    def __init__(self, storage, name, runlist):
        super(Upload, self).__init__(storage, name)
        self.runlist = runlist
        if isinstance(self.runlist, types.DictType):
            return
        elif isinstance(self.runlist, types.StringTypes) and len(self.runlist.strip()) > 0:
            return
        else:
            raise ValueError('Please specify runlist file path')

    @coroutine
    def execute(self):
        runlist = CocaineConfigReader.load(self.runlist)
        with printer('Uploading "%s"', self.name):
            yield self.storage.write('runlists', self.name, runlist, RUNLISTS_TAGS)


class Create(Specific):
    @coroutine
    def execute(self):
        yield Upload(self.storage, self.name, '{}').execute()


class Remove(Specific):
    @coroutine
    def execute(self):
        log.info('Removing "%s"... ', self.name)
        yield self.storage.remove('runlists', self.name)
        log.info('OK')


class Copy(Specific):
    def __init__(self, storage, name, copyname):
        super(Copy, self).__init__(storage, name)
        self.copyname = copyname
        if self.name == self.copyname:
            raise ToolsError("unable to copy an instance to itself")

    @coroutine
    def execute(self):
        log.info('Rename "%s" to "%s"', self.name, self.copyname)
        oldprofile = yield View(self.storage, self.name).execute()
        yield Upload(self.storage, self.copyname, oldprofile).execute()


class Rename(Copy):
    @coroutine
    def execute(self):
        yield super(Rename, self).execute()
        yield Remove(self.storage, self.name).execute()


class AddApplication(Specific):
    def __init__(self, storage, name, app, profile, force=False):
        super(AddApplication, self).__init__(storage, name)
        self.app = app
        self.profile = profile
        self.force = force
        if not self.app:
            raise ValueError('Please specify application name')
        if not self.profile:
            raise ValueError('Please specify profile')

    @coroutine
    def execute(self):
        result = {
            'runlist': self.name,
            'status': 'modified',
            'added': {
                'app': self.app,
                'profile': self.profile,
            }
        }

        runlists = yield List(self.storage).execute()
        if self.force and self.name not in runlists:
            log.debug('Runlist does not exist. Create a new one ...')
            yield Create(self.storage, self.name).execute()
            result['status'] = 'created'

        runlist = yield View(self.storage, name=self.name).execute()
        log.debug('Found runlist: %s', runlist)
        runlist[self.app] = self.profile
        yield Upload(self.storage, name=self.name, runlist=runlist).execute()
        raise gen.Return(result)


class RemoveApplication(Specific):
    def __init__(self, storage, name, app):
        super(RemoveApplication, self).__init__(storage, name)
        self.app = app
        if not self.app:
            raise ValueError('Please specify application name')

    @coroutine
    def execute(self):
        result = {
            'runlist': self.name,
            'app': self.app,
            'status': 'successfully removed',
        }

        runlists = yield List(self.storage).execute()
        if self.name not in runlists:
            log.debug('Runlist does not exist')
            raise ToolsError('Runlist %s is missing', self.name)

        runlist = yield View(self.storage, name=self.name).execute()
        log.debug('Found runlist: %s', runlist)
        if runlist.pop(self.app, None) is None:
            result['status'] = 'the application named {0} is not in runlist'.format(self.app)
        else:
            yield Upload(self.storage, name=self.name, runlist=runlist).execute()
        raise gen.Return(result)



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os

from tornado import gen

from cocaine.exceptions import ServiceError
from cocaine.tools.error import ToolsError


class TracingConfigurator(object):
    def __init__(self, configuration_service, path="/zipkin_sampling"):
        self.configuration_service = configuration_service
        self.path = path

    @gen.coroutine
    def execute(self):
        raise NotImplementedError


class TracingConfigRemove(TracingConfigurator):
    def __init__(self, name, configuration_service, path="/zipkin_sampling"):
        super(TracingConfigRemove, self).__init__(configuration_service, path)
        self.name = name

    @gen.coroutine
    def execute(self):
        abs_node_path = os.path.join(self.path, self.name)
        removed = yield (yield self.configuration_service.remove(abs_node_path, 0)).rx.get()
        if not removed:
            raise ToolsError("the value at %s was not removed" % (abs_node_path))


class TracingConfigView(TracingConfigurator):
    def __init__(self, configuration_service, path="/zipkin_sampling", name=None):
        super(TracingConfigView, self).__init__(configuration_service, path)
        self.name = name

    @gen.coroutine
    def execute(self):
        if self.name:
            res = yield self.specific(self.name)
            raise gen.Return(res)
        else:
            listing_channel = yield self.configuration_service.children_subscribe(self.path)
            _, listing = yield listing_channel.rx.get()
            listing_channel.tx.close()
            res = {}
            for node in listing:
                node_res = yield self.specific(node)
                res[node] = node_res
            raise gen.Return(res)

    @gen.coroutine
    def specific(self, name):
        channel = yield self.configuration_service.get(os.path.join(self.path, name))
        value, version = yield channel.rx.get()
        raise gen.Return({"version": version, "value": value})


def convert_tracing_config_value(value):
    try:
        return float(value)
    except ValueError as err:
        raise ToolsError("value %s must be convertable to float: %s" % (value, err))


class TracingConfigStore(TracingConfigurator):
    def __init__(self, name, value, configuration_service, path="/zipkin_sampling"):
        super(TracingConfigStore, self).__init__(configuration_service, path)
        self.name = name
        self.value = convert_tracing_config_value(value)

    @gen.coroutine
    def execute(self):
        abs_node_path = os.path.join(self.path, self.name)
        try:
            channel = yield self.configuration_service.create(abs_node_path, self.value)
            yield channel.rx.get()
        except ServiceError:
            try:
                version, _ = yield (yield self.configuration_service.get(abs_node_path)).rx.get()
                saved, _ = yield (yield self.configuration_service.put(abs_node_path, self.value, version))
                if not saved:
                    raise ToolsError("the value was not stored to %s" % (abs_node_path))
            except Exception as err:
                raise ToolsError("unable to store value at %s: %s" % (abs_node_path, err))



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import json
import tarfile

import msgpack
from tornado import gen

from cocaine.decorators import coroutine
from cocaine.tools import log

__author__ = 'Evgeny Safronov <division494@gmail.com>'


def isJsonValid(text):
    try:
        json.loads(text)
        return True
    except ValueError:
        return False


def readArchive(filename):
    if not tarfile.is_tarfile(filename):
        raise tarfile.TarError('File "{0}" is not tar file'.format(filename))
    with open(filename, 'rb') as archive:
        return archive.read()


class CocaineConfigReader:
    @classmethod
    def load(cls, context, validate=lambda ctx: None):
        if isinstance(context, dict):
            log.debug('Content specified directly by dict')
            validate(context)
            return msgpack.dumps(context)

        if isJsonValid(context):
            log.debug('Content specified directly by string')
            content = context
        else:
            log.debug('Loading content from file ...')
            with open(context, 'rb') as fh:
                content = fh.read()
        content = json.loads(content)
        validate(content)
        return msgpack.dumps(content)


class Storage(object):
    def __init__(self, storage=None):
        self.storage = storage

    @coroutine
    def execute(self):  # pragma: no cover
        raise NotImplementedError()


class List(Storage):
    """
    Abstract storage action class which main aim is to provide find list action on 'key' and 'tags'.
    For example if key='manifests' and tags=('apps',) this class will try to find applications list
    """
    def __init__(self, key, tags, storage):
        super(List, self).__init__(storage)
        self.key = key
        self.tags = tags

    @coroutine
    def execute(self):
        channel = yield self.storage.find(self.key, self.tags)
        listing = yield channel.rx.get()
        raise gen.Return(listing)


class Specific(Storage):
    def __init__(self, storage, entity, name):
        super(Specific, self).__init__(storage)
        self.name = name
        if not self.name:
            raise ValueError('Please specify {0} name'.format(entity))


class View(Specific):
    def __init__(self, storage, entity, name, collection):
        super(View, self).__init__(storage, entity, name)
        self.collection = collection

    @coroutine
    def execute(self):
        channel = yield self.storage.read(self.collection, self.name)
        value = yield channel.rx.get()
        raise gen.Return(msgpack.loads(value))



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import base64
import json
import os
import tarfile
import StringIO
import urllib

from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.httputil import HTTPHeaders
from tornado.ioloop import IOLoop
from tornado import gen

from cocaine.tools import log
from cocaine.decorators import coroutine
from cocaine.tools.helpers._unix import AsyncUnixHTTPClient
from cocaine.tools.helpers.dockertemplate import dockerchef, dockerpuppet
from cocaine.tools.helpers import JSONUnpacker

__author__ = 'Evgeny Safronov <division494@gmail.com>'

DEFAULT_TIMEOUT = 3600.0
DEFAULT_URL = 'unix://var/run/docker.sock'
DEFAULT_VERSION = '1.15'
REGISTRY_AUTH_HEADER = 'X-Registry-Auth'
DEFAULT_INDEX_URL = 'https://index.docker.io/v1/'


class DockerException(Exception):
    pass


def expand_registry_url(hostname):
    if hostname.startswith('http:') or hostname.startswith('https:'):
        if '/' not in hostname[9:]:
            hostname += '/v1/'
        return hostname
    return 'http://' + hostname + '/v1/'


def resolve_repository_name(fullname):
    if '://' in fullname:
        raise ValueError('repository name can not contain a scheme ({0})'.format(fullname))

    parts = fullname.split('/', 1)
    if '.' not in parts[0] and ':' not in parts[0] and parts[0] != 'localhost':
        return DEFAULT_INDEX_URL, fullname

    if len(parts) < 2:
        raise ValueError('invalid repository name ({0})'.format(fullname))

    if 'index.docker.io' in parts[0]:
        raise ValueError('invalid repository name, try "{0}" instead'.format(parts[1]))

    return expand_registry_url(parts[0]), parts[1]


class Client(object):
    def __init__(self, url=DEFAULT_URL, version=DEFAULT_VERSION, timeout=DEFAULT_TIMEOUT, io_loop=None):
        self.url = url
        self.version = version
        self.timeout = timeout
        self._io_loop = io_loop or IOLoop.current()
        self.config = {
            'url': url,
            'version': version,
            'timeout': timeout,
            'io_loop': io_loop or IOLoop.current()
        }

    def info(self):
        return Info(**self.config).execute()

    def images(self):
        return Images(**self.config).execute()

    def containers(self):
        return Containers(**self.config).execute()

    def build(self, path, tag=None, quiet=False, streaming=None):
        return Build(path, tag, quiet, streaming, **self.config).execute()

    def pull(self, name, auth, streaming=None):
        return Pull(name, auth, streaming, **self.config).execute()

    def tag(self, name, auth, tag, streaming=None):
        return Tag(name, auth, tag, streaming, **self.config).execute()

    def push(self, name, auth, streaming=None):
        return Push(name, auth, streaming, **self.config).execute()


class Action(object):
    def __init__(self, url, version, timeout=DEFAULT_TIMEOUT, io_loop=None):
        self._unix = url.startswith('unix://')

        self._version = version
        self.timeout = timeout
        self._io_loop = io_loop
        if self._unix:
            # url should mimicry to http://
            # to pass an urlscheme check in _HTTPConnetcion.
            # Overriden Unix resolver'll return proper url and AF.
            self._base_url = "http://unixsocket"
            self._http_client = AsyncUnixHTTPClient(self._io_loop, url)
        else:
            self._base_url = '{0}/v{1}'.format(url, version)
            self._http_client = AsyncHTTPClient(self._io_loop)

    def execute(self):
        raise NotImplementedError

    def _make_url(self, path, query=None):
        if query is not None:
            query = dict((k, v) for k, v, in query.items() if v is not None)
            return '{0}{1}?{2}'.format(self._base_url, path, urllib.urlencode(query))
        else:
            return '{0}{1}'.format(self._base_url, path)


class Info(Action):
    @coroutine
    def execute(self):
        response = yield self._http_client.fetch(self._make_url('/info'),
                                                 allow_ipv6=True)
        raise gen.Return(json.loads(response.body))


class Images(Action):
    @coroutine
    def execute(self):
        response = yield self._http_client.fetch(self._make_url('/images/json'),
                                                 allow_ipv6=True)
        raise gen.Return(json.loads(response.body))


class Containers(Action):
    @coroutine
    def execute(self):
        response = yield self._http_client.fetch(self._make_url('/containers/json'),
                                                 allow_ipv6=True)
        raise gen.Return(json.loads(response.body))


class StreamingAction(Action):
    def __init__(self, auth=None, streaming=None, url=DEFAULT_URL,
                 version=DEFAULT_VERSION, timeout=DEFAULT_TIMEOUT, io_loop=None):
        self.auth = auth or {}
        self._streaming = streaming
        self._jsonunpacker = JSONUnpacker()

        self._lasterr = None
        super(StreamingAction, self).__init__(url, version, timeout, io_loop)

    def _prepare_auth_header_value(self):
        username = self.auth.get('username', 'username')
        password = self.auth.get('password', 'password')
        return base64.b64encode('{0}:{1}'.format(username, password))

    def _handle_message(self, message):
        if "stream" in message:
            log.info(message["stream"].rstrip('\n'))
        elif "error" in message:
            error_msg = message["error"].rstrip('\n')
            self._lasterr = DockerException(error_msg)
            log.error(error_msg)

        if self._streaming is not None:
            self._streaming(message)

    def _on_body(self, data):
        self._jsonunpacker.feed(data)
        for i in self._jsonunpacker:
            self._handle_message(i)


class Build(StreamingAction):
    def __init__(self, path, tag=None, quiet=False, *args, **kwargs):
        super(Build, self).__init__(*args, **kwargs)
        self._path = path
        self._tag = tag
        self._quiet = quiet

    @coroutine
    def execute(self):
        headers = None
        body = ''
        remote = None

        if any(map(self._path.startswith, ['http://', 'https://', 'git://', 'github.com/'])):
            log.info('Remote url detected: "%s"', self._path)
            remote = self._path
        else:
            log.info('Local path detected. Creating archive "%s"... ', self._path)
            headers = {'Content-Type': 'application/tar'}
            body = self.make_env(self._path)
            log.info('OK')

        query = {'t': self._tag,
                 'remote': remote,
                 'q': self._quiet}

        url = self._make_url('/build', query)
        log.info('Building "%s"... ', url)
        request = HTTPRequest(url,
                              method='POST',
                              headers=headers, body=body,
                              request_timeout=self.timeout,
                              allow_ipv6=True,
                              streaming_callback=self._on_body)
        try:
            result = yield self._http_client.fetch(request)
            if self._lasterr is not None:
                raise self._lasterr
            log.info('OK')
        except Exception as err:
            log.error('FAIL - %s', err)
            raise err
        else:
            raise gen.Return(result)

    def _tar(self, path):
        stream = StringIO.StringIO()
        try:
            tar = tarfile.open(mode='w', fileobj=stream)
            tar.add(path, arcname='.')
            return stream.getvalue()
        finally:
            stream.close()

    def make_env(self, path):
        if os.path.exists(os.path.join(path, "Dockerfile")):
            log.info("Dockerfile exists")
            return self._tar(path)
        elif os.path.isdir(os.path.join(path, "cookbooks")):
            log.info("Chef has been detected")
            dockfilecontent = dockerchef.generate(basecontainer="ubuntu:precise",
                                                  cookbooks="cookbooks")
        elif os.path.isdir(os.path.join(path, "puppet")):
            log.info("Puppet has been detected")
            if not os.path.exists(os.path.join(path, "puppet/cocaine.pp")):
                raise ValueError("You have to name your own Puppet manifest 'cocaine.pp'")

            dockfilecontent = dockerpuppet.generate(basecontainer="ubuntu:precise")
        else:
            raise ValueError("Please, create Dockerfile or Puppet manifest or Chef recipe")

        log.info("Generate Dockerfile")
        log.info(dockfilecontent)
        with open("Dockerfile", "w") as dockerfile:
            dockerfile.write(dockfilecontent)
        try:
            return self._tar(path)
        finally:
            os.unlink("Dockerfile")


class Push(StreamingAction):
    def __init__(self, name, *args, **kwargs):
        self.name = name
        super(Push, self).__init__(*args, **kwargs)

    @coroutine
    def execute(self):
        url = self._make_url('/images/{0}/push'.format(self.name))
        registry, name = resolve_repository_name(self.name)

        headers = HTTPHeaders()
        headers.add(REGISTRY_AUTH_HEADER, self._prepare_auth_header_value())
        body = ''
        log.info('Pushing "%s" into "%s"... ', name, registry)
        log.debug('Pushing url: %s', url)
        request = HTTPRequest(url, method='POST',
                              headers=headers,
                              body=body,
                              allow_ipv6=True,
                              request_timeout=self.timeout,
                              streaming_callback=self._on_body)
        try:
            result = yield self._http_client.fetch(request)
            if self._lasterr is not None:
                raise self._lasterr
            log.info('OK')
        except Exception as err:
            log.error('FAIL - %s', err)
            raise err

        raise gen.Return(result)


class Pull(StreamingAction):
    def __init__(self, name, *args, **kwargs):
        self.name = name
        super(Pull, self).__init__(*args, **kwargs)

    @coroutine
    def execute(self):
        url = self._make_url('/images/create', query={"fromImage": self.name})
        registry, name = resolve_repository_name(self.name)

        headers = HTTPHeaders()
        headers.add(REGISTRY_AUTH_HEADER, self._prepare_auth_header_value())
        body = ''
        log.info('Pulling "%s" ... ', name)
        request = HTTPRequest(url, method='POST',
                              headers=headers,
                              body=body,
                              allow_ipv6=True,
                              request_timeout=self.timeout,
                              streaming_callback=self._on_body)
        try:
            result = yield self._http_client.fetch(request)
            if self._lasterr is not None:
                raise self._lasterr
            log.info('OK')
        except Exception as err:
            log.error('FAIL - %s', err)
            raise err

        raise gen.Return(result)


class Tag(StreamingAction):
    def __init__(self, name, tag, *args, **kwargs):
        self.name = name
        self.tag = tag
        super(Tag, self).__init__(*args, **kwargs)

    @coroutine
    def execute(self):
        url = self._make_url('/images/{0}/tag'.format(self.name), query={"repo": self.tag})
        registry, name = resolve_repository_name(self.name)

        headers = HTTPHeaders()
        headers.add(REGISTRY_AUTH_HEADER, self._prepare_auth_header_value())
        body = ''
        log.info('Tagging "%s" with "%s"" ... ', name, self.tag)
        request = HTTPRequest(url, method='POST',
                              headers=headers,
                              body=body,
                              allow_ipv6=True,
                              request_timeout=self.timeout,
                              streaming_callback=self._on_body)
        try:
            result = yield self._http_client.fetch(request)
            if self._lasterr is not None:
                raise self._lasterr
            log.info('OK')
        except Exception as err:
            log.error('FAIL - %s', err)
            raise err

        raise gen.Return(result)



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import msgpack

from cocaine.decorators import coroutine

from cocaine.tools.error import ToolsError
from cocaine.tools import actions
from cocaine.tools.actions import CocaineConfigReader, log
from cocaine.tools.tags import GROUPS_TAGS

__author__ = 'EvgenySafronov <division494@gmail.com>'

GROUP_COLLECTION = 'groups'


class MalformedGroup(ValueError):
    pass


class GroupWithZeroTotalWeight(MalformedGroup):
    pass


def validate_routing_group(group):
    accepted_value_types = (int, long)

    if not group:
        raise GroupWithZeroTotalWeight("routing group must not be empty")

    if not sum(group.values()):
        raise GroupWithZeroTotalWeight("routing group must have non-zero total weight")

    if not all(((isinstance(value, accepted_value_types) and value >= 0)
                for value in group.values())):
        raise MalformedGroup("weight must be positive integer value")


class Specific(actions.Specific):
    def __init__(self, storage, name):
        super(Specific, self).__init__(storage, 'group', name)


class List(actions.List):
    def __init__(self, storage):
        super(List, self).__init__(GROUP_COLLECTION, GROUPS_TAGS, storage)


class View(actions.View):
    def __init__(self, storage, name):
        super(View, self).__init__(storage, GROUP_COLLECTION, name, 'groups')


class Create(actions.Specific):
    def __init__(self, storage, name, content):
        super(Create, self).__init__(storage, 'group', name)
        self.content = CocaineConfigReader.load(content, validate=validate_routing_group)

    @coroutine
    def execute(self):
        channel = yield self.storage.write(GROUP_COLLECTION, self.name, self.content, GROUPS_TAGS)
        yield channel.rx.get()


class Remove(actions.Specific):
    def __init__(self, storage, name):
        super(Remove, self).__init__(storage, 'group', name)

    @coroutine
    def execute(self):
        channel = yield self.storage.remove(GROUP_COLLECTION, self.name)
        yield channel.rx.get()


class Copy(Specific):
    def __init__(self, storage, name, copyname):
        super(Copy, self).__init__(storage, name)
        self.copyname = copyname
        if self.name == self.copyname:
            raise ToolsError("unable to copy an instance to itself")

    @coroutine
    def execute(self):
        log.info('Rename "%s" to "%s"', self.name, self.copyname)
        oldprofile = yield View(self.storage, self.name).execute()
        yield Create(self.storage, self.copyname, oldprofile).execute()


class Rename(Copy):
    @coroutine
    def execute(self):
        yield super(Rename, self).execute()
        yield Remove(self.storage, self.name).execute()


class Refresh(actions.Storage):
    def __init__(self, locator, storage, name):
        super(Refresh, self).__init__(storage)
        self.locator = locator
        self.name = name

    @coroutine
    def execute(self):
        if not self.name:
            names = yield List(self.storage).execute()
        else:
            names = [self.name]

        channel = yield self.locator.refresh(names)
        yield channel.rx.get()


class AddApplication(actions.Specific):
    def __init__(self, storage, name, app, weight):
        super(AddApplication, self).__init__(storage, 'group', name)
        self.app = app
        self.weight = int(weight)

    @coroutine
    def execute(self):
        channel = yield self.storage.read(GROUP_COLLECTION, self.name)
        group = yield channel.rx.get()
        group = msgpack.loads(group)
        group[self.app] = self.weight
        validate_routing_group(group)
        channel = yield self.storage.write(GROUP_COLLECTION, self.name, msgpack.dumps(group), GROUPS_TAGS)
        yield channel.rx.get()


class RemoveApplication(actions.Specific):
    def __init__(self, storage, name, app):
        super(RemoveApplication, self).__init__(storage, 'group', name)
        self.app = app

    @coroutine
    def execute(self):
        channel = yield self.storage.read(GROUP_COLLECTION, self.name)
        group = yield channel.rx.get()
        group = msgpack.loads(group)
        if self.app in group:
            del group[self.app]
        validate_routing_group(group)
        channel = yield self.storage.write(GROUP_COLLECTION, self.name, msgpack.dumps(group), GROUPS_TAGS)
        yield channel.rx.get()



#
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function

import contextlib
import sys

from cocaine.tools import log


__author__ = 'EvgenySafronov <division494@gmail.com>'


ENABLE_OUTPUT = False


class Color:
    RESET = '\033[0m'
    OFFSET = 30
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = ['\033[1;%dm' % (OFFSET + id_) for id_ in range(8)]


class Result(object):
    def __init__(self):
        self.value = ''

    def set(self, msg, *args):
        self.value = ' - {0}'.format('{0}{1}{2}'.format(Color.WHITE, msg % args, Color.RESET))

    def __str__(self):
        return str(self.value)


def _print(status, message, color, suffix):
    status = '{0:^6}'.format(status)
    formatted = '[{0}{1}{2}] {3}{4}'.format(color, status, Color.RESET, message, suffix)
    if ENABLE_OUTPUT:
        sys.stdout.write(formatted)
        sys.stdout.flush()
    else:
        log.debug(formatted)


def print_start(message):
    _print('', message, Color.WHITE, '\r')


def print_success(message):
    _print('OK', message, Color.GREEN, '\n')


def print_error(message):
    _print('FAIL', message, Color.RED, '\n')


@contextlib.contextmanager
def printer(message, *args):
    result = Result()
    message = message % args

    try:
        print_start(message)
        yield result.set
        print_success('{0}{1}'.format(message, result))
    except Exception:
        print_error('{0}{1}'.format(message, result))
        raise



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import errno
import json
import msgpack
import time

from tornado.ioloop import IOLoop

from cocaine.exceptions import ChokeEvent, CocaineError
from cocaine.decorators import coroutine
from cocaine.tools import log, interactive
from cocaine.tools.actions import common, app, profile, runlist, crashlog, group, tracing
from cocaine.tools.error import ToolsError


__author__ = 'EvgenySafronov <division494@gmail.com>'


class ToolHandler(object):
    def __init__(self, action):
        self._Action = action

    @coroutine
    def execute(self, **config):
        try:
            action = self._Action(**config)
            result = yield action.execute()
            self._processResult(result)
        except IOError:
            exit(128)
        except (ChokeEvent, StopIteration):
            pass
        except (CocaineError, ToolsError) as err:
            log.error(err)
            exit(128)
        except ValueError as err:
            log.error(err)
            exit(errno.EINVAL)
        except Exception as err:
            log.exception(err)
            exit(128)

    def _processResult(self, result):
        pass


class JsonToolHandler(ToolHandler):
    def _processResult(self, result):
        print(json.dumps(result, sort_keys=True, indent=4))


class PrintToolHandler(ToolHandler):
    def _processResult(self, result):
        print(result)


class CrashlogStatusToolHandler(ToolHandler):
    FORMAT_HEADER = '{0:^20} {1:^10} {2:^26} {3:^36}'
    HEADER = FORMAT_HEADER.format('Application', 'Total', 'Last', 'UUID')
    FORMAT_CONTENT = '{0:<20}|{1:^10}|{2:^26}|{3:^38}'

    def _processResult(self, result):
        if not result:
            print('There are no applications with crashlogs')

        log.info(self.HEADER)
        for appname, (timestamp, time_, uuid), total in sorted(result,
                                                               key=lambda (app, (timestamp, time_, uuid), total): timestamp):
            print(self.FORMAT_CONTENT.format(appname, total, time, uuid))


class CrashlogListToolHandler(ToolHandler):
    FORMAT_HEADER = '{0:^20} {1:^26} {2:^36}'
    HEADER = FORMAT_HEADER.format('Timestamp', 'Time', 'UUID')

    def _processResult(self, result):
        if not result:
            log.info('Crashlog list is empty')
            return

        log.info(self.HEADER)
        for timestamp, caltime, uuid in sorted(crashlog._parseCrashlogs(result), key=lambda (ts, caltime, uuid): ts):
            print(self.FORMAT_HEADER.format(timestamp, caltime, uuid))


class CrashlogViewToolHandler(ToolHandler):
    def _processResult(self, result):
        print('\n'.join(msgpack.loads(result)))


class CallActionCli(ToolHandler):
    def _processResult(self, result):
        requestType = result['request']
        response = result['response']
        if requestType == 'api':
            log.info('Service provides following API:')
            log.info('\n'.join(' - {0}'.format(method) for method in response))
        elif requestType == 'invoke':
            print(response)


NG_ACTIONS = {
    'cluster': JsonToolHandler(common.Cluster),
    'info': JsonToolHandler(common.NodeInfo),
    'locate': JsonToolHandler(common.Locate),
    'routing': JsonToolHandler(common.Routing),

    'app:check': ToolHandler(app.Check),
    'app:list': JsonToolHandler(app.List),
    'app:view': JsonToolHandler(app.View),
    'app:remove': ToolHandler(app.Remove),
    'app:upload': ToolHandler(app.LocalUpload),
    'app:upload-docker': ToolHandler(app.DockerUpload),
    'app:import-docker': ToolHandler(app.DockerImport),
    'app:upload-manual': ToolHandler(app.Upload),
    'app:start': PrintToolHandler(app.Start),
    'app:pause': PrintToolHandler(app.Stop),
    'app:stop': PrintToolHandler(app.Stop),
    'app:restart': PrintToolHandler(app.Restart),

    'profile:copy': ToolHandler(profile.Copy),
    'profile:edit': ToolHandler(interactive.ProfileEditor),
    'profile:list': JsonToolHandler(profile.List),
    'profile:remove': ToolHandler(profile.Remove),
    'profile:rename': ToolHandler(profile.Rename),
    'profile:upload': ToolHandler(profile.Upload),
    'profile:view': JsonToolHandler(profile.View),

    'runlist:add-app': JsonToolHandler(runlist.AddApplication),
    'runlist:create': ToolHandler(runlist.Create),
    'runlist:copy': ToolHandler(runlist.Copy),
    'runlist:edit': ToolHandler(interactive.RunlistEditor),
    'runlist:list': JsonToolHandler(runlist.List),
    'runlist:remove': ToolHandler(runlist.Remove),
    'runlist:remove-app': JsonToolHandler(runlist.RemoveApplication),
    'runlist:rename': ToolHandler(runlist.Rename),
    'runlist:upload': ToolHandler(runlist.Upload),
    'runlist:view': JsonToolHandler(runlist.View),

    'group:app:add': ToolHandler(group.AddApplication),
    'group:app:remove': ToolHandler(group.RemoveApplication),
    'group:create': ToolHandler(group.Create),
    'group:copy': ToolHandler(group.Copy),
    'group:edit': ToolHandler(interactive.GroupEditor),
    'group:list': JsonToolHandler(group.List),
    'group:remove': ToolHandler(group.Remove),
    'group:rename': ToolHandler(group.Rename),
    'group:refresh': ToolHandler(group.Refresh),
    'group:view': JsonToolHandler(group.View),

    'crashlog:status': CrashlogStatusToolHandler(crashlog.Status),
    'crashlog:list': CrashlogListToolHandler(crashlog.List),
    'crashlog:view': CrashlogViewToolHandler(crashlog.View),
    'crashlog:remove': ToolHandler(crashlog.Remove),
    'crashlog:removeall': ToolHandler(crashlog.RemoveAll),
    'crashlog:clean': ToolHandler(crashlog.Clean),
    'crashlog:cleanwhen': ToolHandler(crashlog.CleanRange),

    'tracing:view': JsonToolHandler(tracing.TracingConfigView),
    'tracing:store': ToolHandler(tracing.TracingConfigStore),
    'tracing:remove': ToolHandler(tracing.TracingConfigRemove),
}


class Executor(object):
    """
    This class represents abstract action executor for specified service 'serviceName' and actions pool
    """
    def __init__(self, timeout=None):
        self.timeout = timeout
        self._loop = None

    @property
    def loop(self):
        """Lazy event loop initialization"""
        if not self._loop:
            self._loop = IOLoop.current()
            return self._loop
        return self._loop

    def executeAction(self, actionName, **options):
        """Execute action with specified options.

        Tries to create action from its name and invokes it.

        :param actionName: action name.
        :param options: various action configuration.
        """
        assert actionName in NG_ACTIONS, 'wrong action - {0}'.format(actionName)

        action = NG_ACTIONS[actionName]
        self.loop.run_sync(lambda: action.execute(**options), timeout=self.timeout)



#
#    Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
#    Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
#    Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
#    This file is part of Cocaine.
#
#    Cocaine is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Cocaine is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#


# This is a namespace package
try:  # pragma: no cover
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except ImportError:  # pragma: no cover
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import atexit
import pwd
import time
import os
import sys
from signal import SIGTERM


class Daemon(object):  # pragma: no cover
    def __init__(self, pidfile, userid=None, stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile
        self.userid = userid

    def daemonize(self):
        """Double-fork magic"""
        if self.userid:
            uid = pwd.getpwnam(self.userid).pw_uid
            os.seteuid(uid)
        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as err:
            sys.stderr.write("First fork failed: {0} ({1})\n".format(err.errno, err.strerror))
            sys.exit(1)
        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # Second fork

        try:
            pid = os.fork()
            if pid > 0:
                sys.exit(0)
        except OSError as err:
            sys.stderr.write("Second fork failed: {0} ({1})\n".format(err.errno, err.strerror))
            sys.exit(1)

        sys.stdout.flush()
        sys.stderr.flush()
        si = open(self.stdin, 'r')
        so = open(self.stdout, 'w')
        se = open(self.stderr, 'w')
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write PID file
        atexit.register(self.delpid)
        pid = str(os.getpid())
        open(self.pidfile, 'w').write("%s\n" % pid)

    def delpid(self):
        try:
            os.remove(self.pidfile)
        except Exception:
            pass

    def start(self, *args):
        """
        Start  the daemon
        """

        if os.path.exists(self.pidfile):
            msg = "pidfile exists. Exit.\n"
            sys.stderr.write(msg % self.pidfile)
            sys.exit(1)

        self.daemonize()
        self.run(*args)

    def stop(self):
        """
        Stop daemon.
        """
        try:
            pf = open(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if not pid:
            msg = "pidfile doesn't exist. Exit.\n"
            sys.stderr.write(msg % self.pidfile)
            sys.exit(1)

        # Kill
        try:
            while True:
                os.kill(pid, SIGTERM)
                time.sleep(0.5)
        except OSError as err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
                sys.stdout.write("Process {0} stopped successfully\n".format(pid))
            else:
                print(err)
                sys.exit(1)

    def status(self):
        try:
            with open(self.pidfile, 'r') as pf:
                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if not pid:
            msg = "Stopped\n"
            sys.stderr.write(msg)
            sys.exit(1)

        try:
            os.kill(pid, 0)
        except OSError:
            msg = "Stopped\n"
            sys.stderr.write(msg)
            sys.exit(1)

        sys.stdout.write("Running. PID %d\n" % pid)
        sys.exit(0)

    def restart(self, *args):
        self.stop()
        self.start(*args)

    def run(self, *args):
        pass



#
# Copyright (c) 2013+ Anton Tyurin <noxiouz@yandex.ru>
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division

from collections import defaultdict
import datetime
import fnmatch
import os
import time
import socket

from tornado import gen

from cocaine.decorators import coroutine
from cocaine.tools.error import ToolsError

__author__ = 'Evgeny Safronov <division494@gmail.com>'


def split_by_groups(items, split_by=10):
    return (items[i - split_by:i] for i in xrange(split_by, len(items) + split_by, split_by))


class Node(object):
    def __init__(self, node=None):
        self.node = node

    @coroutine
    def execute(self):
        raise NotImplementedError()


class Locate(object):
    def __init__(self, locator, name):
        self.locator = locator
        if not name:
            raise ValueError("option `name` must be specified")
        self.name = name

    @coroutine
    def execute(self):
        channel = yield self.locator.resolve(self.name)
        endpoints, version, api = yield channel.rx.get()
        result = {
            "endpoints": ["%s:%d" % (addr, port) for addr, port in endpoints],
            "version": version,
            "api": dict((num, method[0]) for num, method in api.items())
        }
        raise gen.Return(result)


class Cluster(object):
    def __init__(self, locator, resolve=True):
        self.locator = locator
        self.resolve = resolve

    @coroutine
    def execute(self):
        channel = yield self.locator.cluster()
        result = yield channel.rx.get()
        if self.resolve:
            raise gen.Return(result)

        converted_result = {}
        for uuid, (addr, port) in result.items():
            try:
                host = socket.gethostbyaddr(addr)[0]
                converted_result[uuid] = [host, port]
            except (socket.gaierror, IOError):
                converted_result[uuid] = [addr, port]

        raise gen.Return(converted_result)


class Routing(object):
    extent = pow(2, 32)

    def __init__(self, locator, name=None):
        self.locator = locator
        self.name = name

    def generate_group(self, body):
        if len(body) == 0:
            return {}
        apps = defaultdict(int)
        # initialize with maximum value
        # from the routing
        prev = body[-1][0] - Routing.extent
        for value, app in body:
            apps[app] += (value - prev)
            prev = value

        output = dict((a, w / Routing.extent) for a, w in apps.items())
        return output

    @coroutine
    def execute(self):
        uid = "tools:%s_%d_%f" % (socket.gethostname(), os.getpid(), time.time())
        channel = yield self.locator.routing(uid, True)
        rings = yield channel.rx.get()
        groups = {}
        if not self.name:
            for name, ring in rings.items():
                groups[name] = self.generate_group(ring)
        elif self.name in rings:
            groups[self.name] = self.generate_group(rings[self.name])
        else:
            raise ToolsError("No such group `%s` in the routing. "
                             "Probably you should refresh the locator" % self.name)

        raise gen.Return(groups)


class NodeInfo(Node):
    def __init__(self, node, locator,
                 name=None, flags=0x1,
                 use_wildcard=False,
                 timeout=5):
        super(NodeInfo, self).__init__(node)
        self.locator = locator
        self._name = name
        self._flags = flags
        self._use_wildcard = use_wildcard
        self.timeout = datetime.timedelta(seconds=timeout)
        self.on_timeout_reply = {
            "error": "info was timeouted",
            "state": "unresponsive",
            "meta": "the error was genearated by tools"
        }

    @coroutine
    def execute(self):
        # name is provided and wildcard is switched off
        # so we use exact match
        if self._name and not self._use_wildcard:
            apps = [self._name]
        else:
            channel = yield self.node.list()
            apps = yield channel.rx.get()
            # wildcard has been already checked
            if self._name:
                apps = fnmatch.filter(apps, self._name)

        result = yield self.info(apps)
        raise gen.Return(result)

    @coroutine
    def info(self, apps):
        infos = {}
        for names in split_by_groups(apps, 25):
            futures = {}
            for app in names:
                res_future = (yield self.node.info(app, self._flags)).rx.get()
                futures[app] = gen.with_timeout(self.timeout, res_future)

            wait_iterator = gen.WaitIterator(**futures)
            while not wait_iterator.done():
                try:
                    info = yield wait_iterator.next()
                    infos[wait_iterator.current_index] = info
                except gen.TimeoutError:
                    infos[wait_iterator.current_index] = self.on_timeout_reply
                except Exception as err:
                    infos[wait_iterator.current_index] = str(err)

        result = {
            'apps': infos,
            'count': len(infos),
        }
        raise gen.Return(result)



#
# Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
# Copyright (c) 2011-2014 Other contributors as noted in the AUTHORS file.
#
# This file is part of Cocaine-tools.
#
# Cocaine is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Cocaine is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import subprocess
import os

__author__ = 'EvgenySafronov <division494@gmail.com>'


class RepositoryDownloadError(Exception):
    pass


class RepositoryDownloader(object):
    def download(self, url, destination):  # pragma: no cover
        raise NotImplementedError


class GitRepositoryDownloader(RepositoryDownloader):
    def __init__(self, stream=None):
        self.stream = stream or open(os.devnull, 'w')

    def download(self, url, destination):
        process = subprocess.Popen(['git', 'clone', url, destination],
                                   stdout=self.stream,
                                   stderr=self.stream)
        process.wait()
        if process.returncode != 0:
            raise RepositoryDownloadError('Cannot download repository from "{0}"'.format(url))



#
#    Copyright (c) 2013+ Evgeny Safronov <division494@gmail.com>
#    Copyright (c) 2011-2013 Other contributors as noted in the AUTHORS file.
#
#    This file is part of Cocaine.
#
#    Cocaine is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    Cocaine is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import abc
import json
import os
import subprocess
import tempfile

from cocaine.decorators import coroutine
from cocaine.tools.actions import group
from cocaine.tools.actions import profile
from cocaine.tools.actions import runlist
from cocaine.tools.helpers.editor import locate_editor
from cocaine.tools.printer import printer


__author__ = 'Evgeny Safronov <division494@gmail.com>'


class BaseEditor(object):
    __metaclass__ = abc.ABCMeta

    EDITORS = ['vim', 'emacs', 'nano']

    @abc.abstractmethod
    def view(self):
        pass

    @abc.abstractmethod
    def upload(self, data):
        pass

    @coroutine
    def execute(self):
        with printer('Loading "%s"', self.name):
            content = yield self.view()

        with printer('Editing "%s"', self.name):
            with tempfile.NamedTemporaryFile(delete=False) as fh:
                name = fh.name
                fh.write(json.dumps(content, indent=4))

            ec = None

            # locate default editor
            default_editor = locate_editor()
            if default_editor is not None:
                default_editor in self.EDITORS and self.EDITORS.remove(default_editor)
                self.EDITORS.insert(0, default_editor)

            for editor in self.EDITORS:
                try:
                    ec = subprocess.call([editor, name])
                    break
                except OSError:
                    continue

            if ec is None:
                raise ValueError('cannot open runlist for editing - any of {0} editors not found'.format(self.EDITORS))
            if ec != 0:
                raise ValueError('editing failed with exit code {0}'.format(ec))

        with open(name) as fh:
            yield self.upload(fh.read())
        # Remove temp file
        os.unlink(name)


class RunlistEditor(BaseEditor, runlist.Specific):
    def __init__(self, *args, **kwargs):
        super(RunlistEditor, self).__init__(*args, **kwargs)

    def view(self):
        return runlist.View(self.storage, self.name).execute()

    def upload(self, data):
        return runlist.Upload(self.storage, self.name, data).execute()


class ProfileEditor(BaseEditor, profile.Specific):
    def __init__(self, *args, **kwargs):
        super(ProfileEditor, self).__init__(*args, **kwargs)

    def view(self):
        return profile.View(self.storage, self.name).execute()

    def upload(self, data):
        return profile.Upload(self.storage, self.name, data).execute()


class GroupEditor(BaseEditor, group.Specific):
    def __init__(self, *args, **kwargs):
        super(GroupEditor, self).__init__(*args, **kwargs)

    def view(self):
        return group.View(self.storage, self.name).execute()

    def upload(self, data):
        return group.Create(self.storage, self.name, data).execute()
